! Author: Lukasz A. Drozd 
! Modified: 8/8/2018
!    
!    This program is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, version 3 of the License.

!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details:
!    <http://www.gnu.org/licenses/>. 
 
module latexTableColumnClass

         implicit none 
                  
         !====================================================================================================================================================================
         !user defined module settings ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

         logical            :: pdflatex=.true. !in pdflatex mode the class compiles tables to pdf files
         logical            :: stealth=.false. !in stealth mode the class erases files with tex codes and only leaves final output 
         

         !end of user settings ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
         !====================================================================================================================================================================
         
         
         integer, parameter :: column_character_type       = 0, column_integer_type   =  1
         integer, parameter :: column_doubleprecision_type = 2, column_undefined_type = -1
 
         type latexTableColumn
            character(len=256)                  :: title        = ""
            character(len=256)                  :: columnformat = "(a)"
            integer                             :: datatype     = column_undefined_type
            real*8, allocatable                 :: doublevalues(:)
            character(len=256)    , allocatable :: charactervalues(:)
            integer               , allocatable :: integervalues(:)
         end type latexTableColumn
         
         interface constructLatexTableColumn
            module procedure :: newRealColumn
            module procedure :: newIntegerColumn
            module procedure :: newCharColumn
            module procedure :: newDPColumn
         end interface  
         
         private ::  newRealColumn, newIntegerColumn, newCharColumn, newDPColumn
         
      contains 
 
      subroutine newRealColumn(tablecolumn,title,columndata,columnformat)  
         implicit none
         type(latexTableColumn) :: tablecolumn
         character(len=*)       :: title, columnformat
         real*4  :: columndata(:) 
 
         tablecolumn % title        = title
         tablecolumn % columnformat = columnformat
         tablecolumn % datatype     = column_doubleprecision_type

         allocate(tablecolumn % doublevalues(size(columndata)))
         tablecolumn % doublevalues = columndata
         
      end subroutine newRealColumn

      subroutine newDPColumn(tablecolumn,title,columndata,columnformat)  
         implicit none
         type(latexTableColumn) :: tablecolumn
         character(len=*)       :: title, columnformat
         real*8 :: columndata(:) 
         
         tablecolumn % title        = title
         tablecolumn % columnformat = columnformat
         tablecolumn % datatype     = column_doubleprecision_type
         
         allocate(tablecolumn % doublevalues(size(columndata)))
         tablecolumn % doublevalues = columndata
         
      end subroutine newDPColumn

      subroutine newIntegerColumn(tablecolumn,title,columndata,columnformat)  
         implicit none
         type(latexTableColumn) :: tablecolumn
         character(len=*)       :: title, columnformat
         integer                :: columndata(:)

         tablecolumn % title        = title
         tablecolumn % columnformat = columnformat
         tablecolumn % datatype     = column_integer_type
         
         allocate(tablecolumn % integervalues(size(columndata)))
         tablecolumn % integervalues = columndata
         
      end subroutine newIntegerColumn

      subroutine newCharColumn(tablecolumn,title,columndata,columnformat )  
         implicit none
         type(latexTableColumn) :: tablecolumn
         character(len=*)       :: title, columnformat
         character(len=*)       :: columndata(:)
 
         tablecolumn % title        = title
         tablecolumn % columnformat = columnformat
         tablecolumn % datatype     = column_character_type
         
         allocate(tablecolumn % charactervalues(size(columndata)))
         tablecolumn % charactervalues = columndata
         
      end subroutine newCharColumn
 
      function numberofrows(tablecolumn) result(n)
         implicit none  
         type(latexTableColumn) :: tablecolumn
         integer                :: n
         
         select case ( tablecolumn % datatype )
            case( column_doubleprecision_type ) 
               n = size(tablecolumn % doublevalues) 
            case( column_character_type ) 
               n = size(tablecolumn % charactervalues) 
            case( column_integer_type ) 
               n = size(tablecolumn % integervalues) 
            case default 
               n = 0
         end select 
     end function numberofrows
      
 end module latexTableColumnClass
 
!=======================================================================
    
Module latexTableMakerClass
 
 use latexTableColumnClass
 implicit none
  
    type columnHeaderLayer
            integer                                       :: level 
            integer                                       :: size 
            character(len=100), dimension(:), allocatable :: headers
            integer, dimension (:), allocatable           :: span
    end type columnHeaderLayer
    
    type rowHeader
            integer             :: row  
            character(len=100)  :: header 
    end type rowHeader
    
    type latexTableMaker 
        
            character(len=256), private  :: entryfmt    = "g12.5"
            character(len=1)  , private  :: separator   = "l" 
            type(columnHeaderLayer), dimension(:), allocatable  :: columnHeaderLayers
            integer, private                                    :: numColumnHeaderLayers=0
            integer, private                                    :: numRowHeaders=0
            type(rowHeader), dimension(:), allocatable          :: rowHeaders
    
    contains
    
               procedure :: addColumnHeaderLayer
               procedure :: addRowBlockHeader 
               procedure :: setTableEntryFormat
               procedure :: tableentryformat
               procedure :: clearColumnHeaderLayers
               procedure :: clearRowBlockHeaders

               procedure, private  :: writeDPTableAsLatex
               procedure, private  :: writeRealTableAsLatex
               procedure, private  :: writeIntegerTableAsLatex
               procedure, private  :: writeTableColumnArrayAsLatex 
               generic             :: writeTableAsLatex => writeDPTableAsLatex,  writeIntegerTableAsLatex, writeTableColumnArrayAsLatex, writeRealTableAsLatex
               
    end type latexTableMaker
 
    contains  
 
      subroutine addColumnHeaderLayer(self, headers, span, level)  
         implicit none
         class(latexTableMaker)                             :: self 
         type(ColumnHeaderLayer), dimension(:), allocatable :: tempheaders
         
         character(len=*) :: headers(:)
         integer          :: span(:)
         integer          :: level 
         
         integer          :: ls, hls
         
         ls=size(headers)
                     
         if(ls .ne. size(span) ) then 
                 print*, 'error in latextableclass: header specification inconsistent!'
                 return
         end if
         
         if( allocated( self % columnHeaderLayers ) ) then 
         
            hls=size (self % columnHeaderLayers)
            allocate(tempheaders(hls))
            tempheaders=self % columnHeaderLayers
             
            deallocate( self % columnHeaderLayers ) 
            allocate (self % columnHeaderLayers(hls+1))
            self % columnHeaderLayers(1:hls)=tempheaders(1:hls)
 
            allocate(self % columnHeaderLayers(hls+1) % headers(1:ls))
            allocate(self % columnHeaderLayers(hls+1) % span(1:ls))
            
            self % columnHeaderLayers(hls+1) % headers(1:ls)= headers(1:ls) 
            self % columnHeaderLayers(hls+1) % span(1:ls)=span(1:ls)
            self % columnHeaderLayers(hls+1) % level=level
            self % columnHeaderLayers(hls+1) % size=ls
            
            self % numColumnHeaderLayers=self % numColumnHeaderLayers+1
      
            deallocate(tempheaders)
            
         else 
             
             allocate (self % columnHeaderLayers(1))
             allocate(self % columnHeaderLayers(1) % headers(1:ls))
             allocate(self % columnHeaderLayers(1) % span(1:ls))

             self % numColumnHeaderLayers=1
             
             self % columnHeaderLayers(1) % headers(1:ls)=headers(1:ls)
             self % columnHeaderLayers(1) % span(1:ls)=span(1:ls)
             self % columnHeaderLayers(1) % level=level
             self % columnHeaderLayers(1) % size=ls
             
         end if 
  
      end subroutine addColumnHeaderLayer

 
      subroutine addRowBlockHeader(self,header,row)  
      
         implicit none
         class(latexTableMaker) :: self 
         type(rowHeader), dimension(:), allocatable :: temprowheaders
         
         character(len=*) :: header 
         integer :: row  
         
         integer :: ls, hls
 
         if( allocated(self % rowHeaders) ) then 
         
            hls=size (self % rowHeaders)
            allocate(temprowheaders(hls))
            temprowheaders=self % rowHeaders
             
            deallocate(self % rowheaders) 
            allocate (self % rowheaders(hls+1))
            self % rowHeaders(1:hls)=temprowheaders(1:hls)
 
            self % rowHeaders(hls+1) % header=trim(header)
            self % rowHeaders(hls+1) % row=row
            
            self % numRowHeaders=self % numrowheaders+1
            
            deallocate(temprowheaders)

         else 

            allocate (self % rowheaders(1)) 
 
            self % rowHeaders(1) % header=trim(header)
            self % rowHeaders(1) % row=row
            
            self % numRowHeaders=1
             
         end if 
  
      end subroutine addRowBlockHeader
      
      subroutine clearColumnHeaderLayers(self)  
      
         implicit none
         class(latexTableMaker) :: self 
 
         deallocate( self % columnHeaderLayers ) 
 
      end subroutine clearColumnHeaderLayers
 
      subroutine clearRowBlockHeaders(self)  
      
         implicit none
         class(latexTableMaker) :: self  
 
         deallocate(self % rowHeaders)  
            
      end subroutine clearRowBlockHeaders
      
      subroutine setTableEntryFormat(self,formatstring)
         implicit none
         class(latexTableMaker)  :: self
         character(len=*)        :: formatstring
         
         self % entryfmt = formatstring
         
      end subroutine setTableEntryFormat
 
      character(len=256) function tableentryformat(self)  
      
         implicit none  
         class(latextablemaker)  :: self
         
         tableentryformat = self % entryfmt
         
      end function tableentryformat
 
      subroutine writeDPTableAsLatex(self, tabledata, headers, caption, label, fileunit )  
      
         implicit none
         class(latexTableMaker)  :: self
         real(kind(1.0d0))       :: tabledata(:,:)
         character(len=*)        :: headers(:)
         character(len=*)        :: caption, label
         integer                 :: fileunit
         integer                 :: nrows,ncolumns,row, column
         character(len=256)      :: fmtstr1,fmtstr2
         
         nrows    = size(tabledata,1)
         ncolumns = size(tabledata,2)
         
         call writeLatexHeader(self,ncolumns,caption,fileunit)
         call writeTableHeader(self,headers,fileunit)
 
!        write DP table data
 
         fmtstr1 = "(*(1x," // trim(self % tableentryformat()) // ",' & ') )"
         fmtstr2 = "(" // trim(self % tableentryformat()) //",a )"
         do row = 1, nrows
            write(fileunit,trim(fmtstr1), advance = "no") (tabledata(row,column), column=1,ncolumns-1)
            write(fileunit,trim(fmtstr2)) tabledata(row,ncolumns)," \\"
         end do
         
         call writeLatexFooter(label,fileunit)

      end subroutine writeDPTableAsLatex
 
      subroutine writeRealTableAsLatex(self, tabledata, headers, caption, label, fileunit)  
      
         implicit none
         class(latexTableMaker)  :: self
         real(kind(1.0e0))       :: tabledata(:,:)
         character(len=*)        :: headers(:)
         character(len=*)        :: caption, label
         integer                 :: fileunit
         integer                 :: nrows,ncolumns,row, column
         character(len=256)      :: fmtstr1,fmtstr2
         
         nrows    = size(tabledata,1)
         ncolumns = size(tabledata,2)
         
         call writeLatexHeader(self,ncolumns,caption,fileunit)
         call writeTableHeader(self,headers,fileunit)
 
!        write real table data
 
         fmtstr1 = "(*(1x," // trim(self % tableentryformat()) // ",' & ') )"
         fmtstr2 = "(" // trim(self % tableentryformat()) //",a )"
         do row = 1, nrows
            write(fileunit,trim(fmtstr1), advance = "no") (tabledata(row,column), column=1,ncolumns-1)
            write(fileunit,trim(fmtstr2)) tabledata(row,ncolumns)," \\"
         end do
         
         call writeLatexFooter(label,fileunit)

      end subroutine writeRealTableAsLatex
 
      subroutine writeIntegerTableAsLatex(self, tabledata, headers, caption, label, fileunit)  
      
         implicit none
         class(latexTableMaker) :: self
         integer                :: tabledata(:,:)
         character(len=*)       :: headers(:)
         character(len=*)       :: caption, label
         integer                :: fileunit
         integer                :: nrows,ncolumns,row, column
         character(len=256)     :: fmtstr1,fmtstr2
         
         nrows    = size(tabledata,1)
         ncolumns = size(tabledata,2)
         
         call writeLatexHeader(self,ncolumns,caption,fileunit)
         call writeTableHeader(self,headers,fileunit)
 
!        write integer table data
 
         fmtstr1 = "(*(1x," // trim(self % tableentryformat()) // ",' & ') )"
         fmtstr2 = "(" // trim(self % tableentryformat()) //",a )"
         do row = 1, nrows
            write(fileunit,trim(fmtstr1), advance = "no") (tabledata(row,column), column=1,ncolumns-1)
            write(fileunit,trim(fmtstr2)) tabledata(row,ncolumns)," \\"
         end do
         
         call writeLatexFooter(label,fileunit)

      end subroutine writeIntegerTableAsLatex
 
      subroutine writeTableColumnArrayAsLatex(self, tablecolumnsarray, title, label)
      
         implicit none
         class(latexTableMaker)            :: self
         type(latexTableColumn)            :: tablecolumnsarray(:)
         character(len=*), optional        :: title
         character(len=*), optional        :: label 
         
         integer                           :: nrows,ncolumns,row, column
         integer                           :: i,k, fileunit, ios
         
         character(len=256), allocatable   :: headers(:)
         character(len=256)                :: tex_file_name 
         
         character(len=1000)                :: caption 
 
!        give tex file a name: 
 
	     if (present(label)) then 	
             
            tex_file_name='Latex-Table-'//trim(label)//'.tex'
            
            if( present( title) ) then 
                caption  = trim(title)
            else
                caption  = 'Table label: ' // trim(label)
            end if 
            
	     else 
            tex_file_name='Latex-Table-'//trim(timestampme())//'.tex"'
            stealth=.true. !switch to stealth mode when no label is provided
            
            if(.not. present( title) ) then 
                caption  = trim(title)
            else
                caption  = 'Table label: ' //trim(timestampme()) 
            end if 
            
         end if
 
!        open file 
 
	    call getunit(fileunit)	!find free unit
        
	    if (fileunit==0) then 
		    print *,'Fatal error in latextableclass: could not get a free fortran unit!'
		    stop
	    end if
	    open (unit=fileunit, file=tex_file_name, status='replace', iostat=ios)	
	    if (ios/=0) then 
		    print *,'Fatal error in latextableclass: could not open data file!'
		    stop
        end if 
 
!       gather header data
 
        ncolumns = size(tablecolumnsarray)
        allocate(headers(ncolumns))
         
        nrows    = 0
        do k = 1, ncolumns
            nrows = max(nrows,numberofrows(tablecolumn = tablecolumnsarray(k)))
            headers(k) = tablecolumnsarray(k) % title
        end do  
 
        call writeLatexHeader( self, ncolumns, caption, fileunit)
        call writeTableHeader( self, headers, fileunit)
 
!       write table data to columns
 
        do row = 1, nrows
 
               if(  self % numRowHeaders >0 .and. row==1) then 
                   do i=1,  self % numRowHeaders 
                       if( row == self % rowHeaders(i) % row  ) then
                            write(fileunit,"(a)" ) '{\it '// trim(self % rowHeaders(i) % header)//'}' // ' \\ '
                       end if 
                   end do 
               end if 
               
               do column = 1, ncolumns
                  if( row > numberofrows(tablecolumn = tablecolumnsarray(column)) ) then
                    write(fileunit,"(a)",advance = "no") " "
                  else
                     select case ( tablecolumnsarray(column) % datatype )
                        case( column_integer_type )
                           write(fileunit,trim(tablecolumnsarray(column) % columnformat),advance = "no") &
                                 tablecolumnsarray(column) % integervalues(row) 
                        case( column_character_type )
                            
                           if(column ==1 .and.  self % numRowHeaders >0 )  write(fileunit,'(a)',advance = "no") '\qquad '
                            
                           write(fileunit,trim(tablecolumnsarray(column) % columnformat),advance = "no") &
                                 trim(tablecolumnsarray(column) % charactervalues(row)) 
                           
                        case( column_doubleprecision_type )
                           write(fileunit,trim(tablecolumnsarray(column) % columnformat),advance = "no") &
                                 tablecolumnsarray(column) % doublevalues(row)
                        case default 
                           write(fileunit,*) " "
                     end select 
                  end if
                  
                  if(column < ncolumns) write(fileunit,"(a)", advance = "no") " & " 

                  if( column==ncolumns) then 
                       write(fileunit,"(a)") " \\" 
                       if ( self % numRowHeaders>0 .and. row>1) then
                           do i=1, self % numRowHeaders
                               if( row ==  self % rowHeaders(i) % row ) then
                                    write(fileunit,"(a)") "\midrule "
                                    write(fileunit,"(a)" ) '{\it '// trim(self % rowHeaders(i) % header) // '} \\' 
                               end if 
                           end do   
                       end if
                   end if     
               end do
        end do
        
        deallocate(headers)
 
!       write footer
 
        call writeLatexFooter( label, fileunit )
       
        close ( unit = fileunit )
         
!       pdf Latex the table
         
        if( pdflatex==.true. )  call pdfLatexTable ( tex_file_name, stealth )

      end subroutine writetablecolumnarrayaslatex
      
      subroutine pdfLatexTable( tex_file, stealth ) !issue command to the system to compile tex file using pdflatex

	    implicit none 
        character (len = *), intent(in) :: tex_file     
        logical, intent (in)            :: stealth  
    
        character (len=100)             :: command
	    integer                         :: exitstat, status
     
        !*@#$ note !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
        !platform dependent: modify appropriately to ensure compatibility 
        print*, '' 
        print*, ' -------------------------------------------------' 
        print*, ' LatexTableClass: Writing table ', trim(tex_file), ' to pdf... ' 
        print*, ' -------------------------------------------------' 
        print*, '' 
        write (command, *) 'pdflatex -synctex=1 -interaction=nonstopmode '//trim(tex_file) 
    
        !*comment if compiler does not support system level commands
	    call execute_command_line(trim(command), .true., exitstat, status)	  	!specit to newer intel compilers
        !*
    
	    if (status.ne.0)  stop 'Fatal error in LatexTableClass: command issued to system failed!'

       !clean up in stealth mode
        if(stealth==.true.) then
 
            write (command, *) 'del ' // trim (tex_file) !delete files
            
            !*comment if compiler does not support system level commands
	        call execute_command_line(trim(command), .true., exitstat, status)	  	!specit to newer intel compilers
            !*
        
            if (status.ne.0)  stop 'Fatal error in LatexTableClass: command issued to system failed! '
 
        end if 
 
       !*@#$ end note !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 
     end subroutine pdfLatexTable
 
     subroutine writeLatexHeader(self,numcols,caption,iunit)  
     
             implicit none  
             class(latexTableMaker)  :: self
         
             integer                 :: iunit
             character(len=*)        :: caption
             integer                 :: m, previouslevel, level0, k,i, j, n, numcols, header, level, positionm1, position
          
             write(iunit,'(a)') '\documentclass[12pt,a4paper]{report}'
             
             write(iunit,'(a)') '\usepackage{amsmath}'
             write(iunit,'(a)') '\usepackage{booktabs}'  

             write(iunit,'(a)') '\begin{document}'

             write(iunit,'(a)') '\newcommand{\ra}[1]{\renewcommand{\arraystretch}{#1}}'
          
             write(iunit,'(a)') '\begin{table*}'
             write(iunit,'(a)') '\ra{1.3}'
             write(iunit,'(a)') '\caption{'//trim(caption)//'}'
             write(iunit,'(a)') '\begin{center}'
             write(iunit,'(a)',advance = "no") '\begin{tabular}{@{}'
             write(iunit,"(*(a2) )",advance = "no") (trim(self % separator), j=1,numcols)
             write(iunit,'(a)') '@{}}\toprule'

             n=self % numColumnHeaderLayers

             do j=1,n
                 level0=-1 !reset max level threshold
                 do m=1,n !find next layer in hierarchy (highest upon first entry)
                     level = self % columnHeaderLayers(m) % level 
                     if(level < previouslevel .or. j==1) then
                         if( level > level0 ) then 
                            level0=level; header = m
                         end if 
                     end if 
                 end do
             
                 previouslevel =  self % columnHeaderLayers(header) % level 
 
                 level=self % columnHeaderLayers(header) % level
                 k    =self % columnHeaderLayers(header) % size
         
                 do i=1,k
                     write(iunit,"(a)",advance = "no") " & "
                     if(self%columnHeaderLayers(header) % span(i)>1) then 
                         write(iunit,"(a, i2, a)",advance = "no")  " \multicolumn{",self%columnHeaderLayers(header) % span(i),"}{c} "
                     end if 
                     write(iunit,"(a, i2, a)",advance = "no")  " {"//trim(self%columnHeaderLayers(header) % headers(i))//"} " 
                 end do
                 write(iunit,"(a)",advance = "yes") " \\ "
             
                 positionm1 = 2
                 do i=1,k
                     position = positionm1 + self%columnHeaderLayers(header) % span(i)-1
                     if( self%columnHeaderLayers(header) % span(i)-1 > 0) then 
                         write(iunit,"(a, i3, a, i3, a)",advance = "no") " \cmidrule{",positionm1,"-",position,"} " 
                     end if 
                 positionm1=position+1
                 end do 
             
                 write(iunit,"(a)",advance = "yes") ""
             
             end do 
     end subroutine writeLatexHeader

     subroutine writeLatexFooter(label,fileunit)  
     
             implicit none  
             integer           :: fileunit
             character(len=*)  :: label
         
             write(fileunit,*) "\bottomrule"
             write(fileunit,*) "\end{tabular}"
             write(fileunit,*) "\end{center}"
             write(fileunit,*) "\label{tab:",trim(label),"}"
             write(fileunit,*) "\end{table*}"
             write(fileunit,*) "\end{document}"
         
     end subroutine writeLatexFooter

     subroutine writeTableHeader(self,headers,fileunit)  
     
             implicit none  
             class(latexTableMaker) :: self
         
             character(len=*)       :: headers(:)
             integer                :: fileunit
             integer                :: ncolumns, j
         
             ncolumns = size(headers)
          
             write(fileunit,"(*(1x,a,' & ') )",advance="no") ( trim(headers(j)), j=1,ncolumns-1)
             write(fileunit,"(a)") trim(headers(ncolumns))," \\ "
             write(fileunit,*) "\midrule"
         
     end subroutine writeTableHeader
      
     function timestampme() result(stamp) !creates a string with current date and time 
     
        implicit none
        character(len=8)  :: date
        character(len=10) :: time
        character(len=33) :: stamp 
    
	    call date_and_time(date,time)
	    stamp= 'date_'//date(7:8)//'-'//date(5:6)//'-'//date(1:4)//'_time_'//time(1:2)//'-'//time(3:4)//'-'//time(5:10)
        
     end function timestampme
 
     subroutine getunit(iunit) !return file unit handle
     
	    implicit none
        
        integer, intent(out) :: iunit
	    integer i, ios, lopen
 
	    iunit=0
	    do i=1,99
		    if (i/= 5 .and. i/=6) then	
			    inquire (unit=i, opened=lopen, iostat=ios)
			    if (ios==0) then
				    if (.not.lopen) then
					    iunit=i
					    return
				    end if
			    end if
		
		    end if
	    end do	
        
    end subroutine getunit
      
end module latexTableMakerClass
